#include <stdio.h>
#include <string.h>		/* memcpy */
#include <stdlib.h>		/* malloc */
#include <sys/queue.h>	/* SLIST, singly linked lists (man 3 queue) */

#define ARRAY_SIZE(a) (sizeof(a) / sizeof((a)[0]))
#define MAX_BP_COMBINATIONS (0x01 << 5)

/* This will come from struct flashchip total_size */
#define FLASHCHIP_SIZE 1024	/* kB */

/* Feel free to skip reading throught the following two macros */
#define BP_CHECK_X(n, m) \
wp_n_new = malloc(sizeof(struct wp)); \
	memcpy(wp_n_new, wp_n_tmp, sizeof(struct wp)); \
	wp_n_new->bp##n = (m); \
	SLIST_INSERT_HEAD(&wp_head_proc, wp_n_new, entries)

#define BP_CHECK(n) \
if (wp_n_tmp->bp##n == X) { \
	BP_CHECK_X(n, 0); \
	BP_CHECK_X(n, 1); \
	j = 1; \
}

enum bit_state {
	X 	= -1,
	OFF = 0,
	ON 	= 1
};

static struct range {
	unsigned int start;
	unsigned int len;	/* in kB */
} **range;
/* Declare a global (scope limited to the file; returned 
   on calling wp_init) range array to hold processed BP table 
 */

/* TODO : More apt to rename to struct wp_config or wp_table or something.
   Should be a part of aonther struct which is a part of struct flashchip.
   Should hold a pointer to the BP table (8 bytes for pointer vs 896 bytes
   for array, although pointer will increase complexity but chips with 
   same density are very likely to have similar config).

   struct flashchip {
     ...
     struct statusreg {
	   struct status_register_layout *layout[MAX_STATUS_REGISTERS];
	   ...
     } statusreg;
     struct wp {
	   struct wp_config {
	     struct range range;
	     enum bit_state bp0;
	     ...
	   } **wp_config;

	   struct range (*bp_to_range)(struct flashchip *flash ...);
	   int (*rangeto_bp)(struct flashchip *flash ...);
	   ...
     } wp;
     ...
   }

   Note that size of wp_config[] should be 32. It doesn't matter if fewer
   entries are made because allocation initializes with zeroes by default
   and BP bitfield 0x00 corresponds to NONE protection (that is how flashrom
   disables block protection as of now!)
 */
struct wp {
	struct range range;
	/* This order is deliberately arranged so that while initializing,
	   if a chip doesn't have BP4, then simply drop it, like - { { 0x0f0000, 8 }, 1, 0, 1, 0 } 
	   For chips with TB bit and (say) 3 BP bits (BP0, BP1, BP2), TB works just like BP3 (of course 
	   BP4 can be dropped like just explained above). Same goes for SEC bit.
	 */
	enum bit_state bp0;
	enum bit_state bp1;
	enum bit_state bp2;
	enum bit_state bp3;
	enum bit_state bp4;

	/* This is required by the SLIST macros */
	SLIST_ENTRY(wp) entries;
};

/* Transform the BP table (struct wp) corresponding to flashchip into 
   a range array (struct range []) indexed by the BP bitfield 
   (e.g. range[0x01] = {0x0f0000, 64}). The range array is static and
   global, reference to it is returned.
   TODO : First parameter should be struct flashchip *flash
 */
struct range **wp_init(struct wp wp_ranges[]) {
	if (range == NULL) {
		range = malloc(MAX_BP_COMBINATIONS * sizeof(struct range));
	}
	int j;
	struct wp *wp_n_tmp, *wp_n_new;

	/* Initialize singly linked lists */
	SLIST_HEAD(wp_head_proc, wp) wp_head_proc = SLIST_HEAD_INITIALIZER(wp_head_proc);
	SLIST_HEAD(wp_head_done, wp) wp_head_done = SLIST_HEAD_INITIALIZER(wp_head_done);
	SLIST_INIT(&wp_head_proc);
	SLIST_INIT(&wp_head_done);

	/* Add entries to proc list - to be processed */
	for (int i = 0; i < 32; i++) {
		SLIST_INSERT_HEAD(&wp_head_proc, &wp_ranges[i], entries);
	}

	/* Resolve all dont-care combinations - [X, 0, 1] becomes [[0, 0, 1], [1, 0, 1]] */
	while (!SLIST_EMPTY(&wp_head_proc)) {
		wp_n_tmp = SLIST_FIRST(&wp_head_proc);
		SLIST_REMOVE_HEAD(&wp_head_proc, entries);
		j = 0;
		/* TODO : In all the if conditions, need to also check if corresponding BP bit
		   even exists for the flashchip
		   (Please refer to the macro above to make sense of comment in context. Please
		   pardon this obfuscation, the macro is defined only so that I would have
		   to type lesser!)
		 */
		BP_CHECK(4)
		else BP_CHECK(3)
		else BP_CHECK(2) 
		else BP_CHECK(1) 
		else BP_CHECK(0)
		if (j == 0) {
			SLIST_INSERT_HEAD(&wp_head_done, wp_n_tmp, entries);
		}
	}

	/* Populate the range array */
	SLIST_FOREACH(wp_n_tmp, &wp_head_done, entries) {
		*(range + (wp_n_tmp->bp4 << 4 | wp_n_tmp->bp3 << 3 | wp_n_tmp->bp2 << 2 | wp_n_tmp->bp1 << 1 | wp_n_tmp->bp0)) = &wp_n_tmp->range;
	}

	return range;
}

/* CMP is Complement Protect, flips range, i.e., protection will now work
   for complemented range.
   TODO : Ideally this should take only struct flashchip *flash as parameter.
   There should be a routine to look up the status register layout whether this
   chip has a CMP bit, then read it and continue.
 */
struct range *bp_to_range(unsigned int bp_bitfield, int CMP) {
	if (CMP) {
		struct range *cmp_range = malloc(sizeof(struct range));
		memcpy(cmp_range, *(range + bp_bitfield), sizeof(struct range));
		if (cmp_range->start) {
			cmp_range->start = 0;
		} else {
			cmp_range->start = cmp_range->len * 1024;
		}
		/* In place of FLASHCHIPS_SIZE, size will come from flash->total_size */
		cmp_range->len = FLASHCHIP_SIZE - cmp_range->len;

		return cmp_range;
	} else {
		return *(range + bp_bitfield);
	}
}

/* TODO : Like the previous function, this should also take struct flashchip *flash
   as argument along with start and len. The same CMP bit discussion above applies here
   as well. 
 */
int range_to_bp(unsigned int start, unsigned int len, int CMP) {
	for (unsigned int bp = 0; bp < 32; bp++)
		if (start == bp_to_range(bp, CMP)->start && len == bp_to_range(bp, CMP)->len)
			return bp;
	return -1;
}

/* W25Q80BL, W25Q80DV, W25Q80DL, W25Q80BW, W25Q80EW,
   GD25LQ80B, GD25VQ80C, GD25Q80C 
 */
struct wp w25q_gd25_ranges[32] = {
	/* { struct range }, BP0, BP1, BP2, BP3, BP4 
	   If a chip were to not have BP4 or BP4 and BP3,
	   then simply dropping them of would do, like -
	   { { 0x0f0000, 64 }, 1, 0, 0 },
	 */
	{ { 0x000000, 0 }, 0, 0, 0, X, X },
	{ { 0x0f0000, 64 }, 1, 0, 0, 0, 0 },
	{ { 0x0e0000, 128 }, 0, 1, 0, 0, 0 },
	{ { 0x0c0000, 256 }, 1, 1, 0, 0, 0 },
	{ { 0x080000, 512 }, 0, 0, 1, 0, 0 },

	{ { 0x000000, 64 }, 1, 0, 0, 1, 0 },
	{ { 0x000000, 128 }, 0, 1, 0, 1, 0 },
	{ { 0x000000, 256 }, 1, 1, 0, 1, 0 },
	{ { 0x000000, 512 }, 0, 0, 1, 1, 0 },
	{ { 0x000000, 1024 }, 1, 0, 1, X, 0 },
	{ { 0x000000, 1024 }, X, 1, 1, X, X },

	{ { 0x0ff000, 4 }, 1, 0, 0, 0, 1 },
	{ { 0x0fe000, 8 }, 0, 1, 0, 0, 1 },
	{ { 0x0fc000, 16 }, 1, 1, 0, 0, 1 },
	{ { 0x0f8000, 32 }, X, 0, 1, 0, 1 },

	{ { 0x000000, 4 }, 1, 0, 0, 1, 1 },
	{ { 0x000000, 8 }, 0, 1, 0, 1, 1 },
	{ { 0x000000, 16 }, 1, 1, 0, 1, 1 },
	{ { 0x000000, 32 }, X, 0, 1, 1, 1 },
};

/*
This representation seems more difficult to represent that the corresponding
one just above, but it makes the code far easier than its counterpart.
Some entries are skipped (those aren't holes!) because default initialization
to zeroes is what we want anyway for those cases.

struct range w25q_gd25_range[32] = {
	[0x00] = {0x000000,    0},
	[0x01] = {0x0f0000,   64},
	[0x02] = {0x0e0000,  128},
	[0x03] = {0x0c0000,  256},
	[0x04] = {0x080000,  512},
	[0x11] = {0x0ff000,    4},
	[0x12] = {0x0fe000,    8},
	[0x13] = {0x0fc000,   16},
	[0x14] = {0x0f8000,   32},
	[0x15] = {0x0f8000,   32},

	[0x05] = {0x000000, 1024},
	[0x06] = {0x000000, 1024},
	[0x07] = {0x000000, 1024},
	[0x0d] = {0x000000, 1024},
	[0x0e] = {0x000000, 1024},
	[0x0f] = {0x000000, 1024},
	[0x16] = {0x000000, 1024},
	[0x17] = {0x000000, 1024},
	[0x1e] = {0x000000, 1024},
	[0x1f] = {0x000000, 1024},

	[0x09] = {0x000000,   64},
	[0x0a] = {0x000000,  128},
	[0x0b] = {0x000000,  256},
	[0x0c] = {0x000000,  512},
	[0x19] = {0x000000,    4},
	[0x1a] = {0x000000,    8},
	[0x1b] = {0x000000,   16},
	[0x1c] = {0x000000,   32},
	[0x1d] = {0x000000,   32},
};
*/

/* EN25QH32 */
struct range en25_range_array[32] = {
	{ 0x000000, 0 },
	{ 0x03f000, 64 },
	{ 0x3e0000, 128 },
	{ 0x3c0000, 256 },
	{ 0x380000, 512 },
	{ 0x300000, 1024 },
	{ 0x200000, 2 * 1024 },
	{ 0x000000, 4 * 1024 },

	{ 0x000000, 0 },
	{ 0x000000, 64 },
	{ 0x000000, 128 },
	{ 0x000000, 256 },
	{ 0x000000, 512 },
	{ 0x000000, 1024 },
	{ 0x000000, 2 * 1024 },
	{ 0x000000, 4 * 1024 },

	/* Repeat what is above because this
	   chip doesn't have a BP4 so BP4 is X 
	 */
	{ 0x000000, 0 },
	{ 0x03f000, 64 },
	{ 0x3e0000, 128 },
	{ 0x3c0000, 256 },
	{ 0x380000, 512 },
	{ 0x300000, 1024 },
	{ 0x200000, 2 * 1024 },
	{ 0x000000, 4 * 1024 },

	{ 0x000000, 0 },
	{ 0x000000, 64 },
	{ 0x000000, 128 },
	{ 0x000000, 256 },
	{ 0x000000, 512 },
	{ 0x000000, 1024 },
	{ 0x000000, 2 * 1024 },
	{ 0x000000, 4 * 1024 }
};

int main(int argc, const char **argv) {
	wp_init(w25q_gd25_ranges);

	printf("CMP=0\n");
	for (int i = 0; i < 32; ++i)
		printf("range[0x%02x]={0x%06x, %4dkB}\n", i, bp_to_range(i, 0)->start, bp_to_range(i, 0)->len);
	printf("\n");

	printf("CMP=1\n");
	for (int i = 0; i < 32; ++i)
		printf("range[0x%02x]={0x%06x, %4dkB}\n", i, bp_to_range(i, 1)->start, bp_to_range(i, 1)->len);

	unsigned int start = 0x000000, len = 1020, CMP = 1;
	printf("{0x%06x, %4dkB} : bp=0x%02x\n", start, len, range_to_bp(start, len, CMP));
	printf("\n");

	*range = en25_range_array;
	for (int i = 0; i < 32; ++i)
		printf("range[0x%02x]={0x%06x, %4dkB}\n", i, bp_to_range(i, 0)->start, bp_to_range(i, 0)->len);

	return 0;
}
